@@ -19,6 +19,32 @@
  * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  */
 
+/*
+ * Copyright (C) 1998 by the FundsXpress, INC.
+ * 
+ * All rights reserved.
+ * 
+ * Export of this software from the United States of America may require
+ * a specific license from the United States Government.  It is the
+ * responsibility of any person or organization contemplating export to
+ * obtain such a license before exporting.
+ * 
+ * WITHIN THAT CONSTRAINT, permission to use, copy, modify, and
+ * distribute this software and its documentation for any purpose and
+ * without fee is hereby granted, provided that the above copyright
+ * notice appear in all copies and that both that copyright notice and
+ * this permission notice appear in supporting documentation, and that
+ * the name of FundsXpress. not be used in advertising or publicity pertaining
+ * to distribution of the software without specific, written prior
+ * permission.  FundsXpress makes no representations about the suitability of
+ * this software for any purpose.  It is provided "as is" without express
+ * or implied warranty.
+ * 
+ * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
+ * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
+ */
+
 #ifndef lint
 char copyright[] =
   "@(#) Copyright (c) 1983 The Regents of the University of California.\n\
@@ -61,6 +87,11 @@ char copyright[] =
 #include <setjmp.h>
 #include <netdb.h>
      
+#ifdef HAVE_SYS_FILIO_H
+/* Solaris needs <sys/filio.h> for FIONREAD */
+#include <sys/filio.h>
+#endif
+
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
@@ -129,29 +160,15 @@ char copyright[] =
 #endif
 
 
-#ifndef roundup
-#define roundup(x,y) ((((x)+(y)-1)/(y))*(y))
-#endif
-
 #ifdef KERBEROS
-#include "krb5.h"
-#include "com_err.h"
+#include <krb5.h>
+#include <com_err.h>
 #include "defines.h"
+#ifdef KRB5_KRB4_COMPAT
+#include <kerberosIV/krb.h>
+#endif
      
 #define RLOGIN_BUFSIZ 5120
-     /*
- * Note that the encrypted rlogin packets take the form of a four-byte
- *length followed by encrypted data.  On writing the data out, a significant
- * performance penalty is suffered (at least one RTT per character, two if we
- * are waiting for a shell to echo) by writing the data separately from the 
- * length.  So, unlike the input buffer, which just contains the output
- * data, the output buffer represents the entire packet.
- */
-
-char des_inbuf[2*RLOGIN_BUFSIZ];       /* needs to be > largest read size */
-char des_outpkt[2*RLOGIN_BUFSIZ+4];      /* needs to be > largest write size */
-krb5_data desinbuf,desoutbuf;
-krb5_encrypt_block eblock;      /* eblock for encrypt/decrypt */
 
 void try_normal();
 char *krb_realm = (char *)0;
@@ -160,15 +177,18 @@ int fflag = 0, Fflag = 0;
 krb5_creds *cred;
 struct sockaddr_in local, foreign;
 krb5_context bsd_context;
+krb5_auth_context auth_context;
+
+#ifdef KRB5_KRB4_COMPAT
+Key_schedule v4_schedule;
+CREDENTIALS v4_cred;
+#endif
 
 #ifndef UCB_RLOGIN
 #define UCB_RLOGIN      "/usr/ucb/rlogin"
 #endif
 
 #include "rpaths.h"
-#else /* !KERBEROS */
-#define des_read read
-#define des_write write
 #endif /* KERBEROS */
 
 # ifndef TIOCPKT_WINDOW
@@ -187,6 +207,7 @@ char	*getenv();
 
 char	*name;
 int 	rem = -1;		/* Remote socket fd */
+int	do_inband = 0;
 char	cmdchar = '~';
 int	eight = 1;		/* Default to 8 bit transmission */
 int	no_local_escape = 0;
@@ -206,6 +227,18 @@ char	*speeds[] =
     "900", "1200", "1800", "2400", "3600", "4800", "7200", "9600",
     "19200", "38400", "EXTA", "EXTB" };
 #else
+/* Solaris note: There are higher values we could use.  But Casper Dik
+   <Casper.Dik@Holland.Sun.Com> mentions in article
+   <casper.938167062@uk-usenet.uk.sun.com> in comp.protocols.kerberos
+   on 1999-09-24 some problems in sending higher values to remote
+   systems (for non-Kerberos rlogind?).  So let's stick with this
+   list.  Even if our current klogind doesn't have the problems, older
+   versions are likely to.
+
+   Daniel S. Riley <dsr@mail.lns.cornell.edu> gives 57600, 76800,
+   115200, 153600, 230400, 307200, 460800 as the higher values.
+   (article <sh6711s713.fsf@lnscu4.lns.cornell.edu> in
+   comp.protocols.kerberos, 1999-09-23) */
 char    *speeds[] =
 { "0", "50", "75", "110", "134", "150", "200", "300",
     "600", "1200", "1800", "2400", "4800", "9600", "19200", "38400" };
@@ -231,6 +264,7 @@ char	*host=0;			/* external, so it can be
 					   reached from confirm_death() */
 
 krb5_sigtype	sigwinch KRB5_PROTOTYPE((int));
+int server_message KRB5_PROTOTYPE((int));
 void oob KRB5_PROTOTYPE((void));
 krb5_sigtype	lostpeer KRB5_PROTOTYPE((int));
 #if __STDC__
@@ -338,8 +372,13 @@ main(argc, argv)
     int sock;
     krb5_flags authopts;
     krb5_error_code status;
+#ifdef KRB5_KRB4_COMPAT
+    KTEXT_ST v4_ticket;
+    MSG_DAT v4_msg_data;
 #endif
-    int debug_port = 0;
+#endif
+    int port, debug_port = 0;
+    enum kcmd_proto kcmd_proto = KCMD_PROTOCOL_COMPAT_HACK;
    
     memset(&defaultservent, 0, sizeof(struct servent));
     if (strrchr(argv[0], '/'))
@@ -358,11 +397,15 @@ main(argc, argv)
 
     if (argc > 0 && !strcmp(*argv, "-D")) {
 	argv++; argc--;
+	if (*argv == NULL) {
+	    fprintf (stderr,
+		     "rlogin: -D flag must be followed by the debug port.\n");
+	    exit (1);
+	}
 	debug_port = htons(atoi(*argv));
 	argv++; argc--;
 	goto another;
     }
-
     if (argc > 0 && !strcmp(*argv, "-d")) {
 	argv++, argc--;
 	options |= SO_DEBUG;
@@ -461,6 +504,16 @@ main(argc, argv)
 	argv++, argc--;
 	goto another;
     }
+    if (argc > 0 && !strcmp(*argv, "-PO")) {
+	kcmd_proto = KCMD_OLD_PROTOCOL;
+	argv++, argc--;
+	goto another;
+    }
+    if (argc > 0 && !strcmp(*argv, "-PN")) {
+	kcmd_proto = KCMD_NEW_PROTOCOL;
+	argv++, argc--;
+	goto another;
+    }
 #endif /* KERBEROS */
     if (host == 0)
       goto usage;
@@ -477,12 +530,12 @@ main(argc, argv)
 	    com_err(argv[0], status, "while initializing krb5");
 	    exit(1);
     }
-    desinbuf.data = des_inbuf;
-    desoutbuf.data = des_outpkt+4;	/* Set up des buffers */
 #endif
 
 
-    if(debug_port == 0) {
+    if (debug_port)
+      port = debug_port;
+    else {
 #ifdef KERBEROS
     /*
      * if there is an entry in /etc/services for Kerberos login,
@@ -505,38 +558,38 @@ main(argc, argv)
       }
 #endif /* KERBEROS */
 
-      debug_port = sp->s_port;
+      port = sp->s_port;
     }
 
 
     if (cp == (char *) NULL) cp = getenv("TERM");
-    if (cp)
-      (void) strcpy(term, cp);
+    if (cp) {
+      (void) strncpy(term, cp, sizeof (term));
+      term[sizeof (term) - 1] = '\0';
+    }
 #ifdef POSIX_TERMIOS
 	if (tcgetattr(0, &ttyb) == 0) {
 		int ospeed = cfgetospeed (&ttyb);
 
-		(void) strcat(term, "/");
+                term[sizeof(term) - 1] = '\0';
+		(void) strncat(term, "/", sizeof(term) - 1 - strlen(term));
 		if (ospeed >= 50)
 			/* On some systems, ospeed is the baud rate itself,
 			   not a table index.  */
 			sprintf (term + strlen (term), "%d", ospeed);
+		else if (ospeed >= sizeof(speeds)/sizeof(char*))
+			/* Past end of table, but not high enough to
+			   look like a real speed.  */
+			(void) strncat (term, speeds[sizeof(speeds)/sizeof(char*) - 1], sizeof(term) - 1 - strlen(term));
 		else {
-			(void) strcat(term, speeds[ospeed]);
-#if 0
-			/* XXX - Not used, since the above code was
-			 * not ifdef'd and it relied on cfget... */
-
-			/* some "posix" systems don't have cfget...
-			 * so used CBAUD if it's there */
-			(void) strcat(term, speeds[ttyb.c_cflag & CBAUD]);
-#endif
+			(void) strncat(term, speeds[ospeed], sizeof(term) - 1 - strlen(term));
 		}
+                term[sizeof (term) - 1] = '\0';
 	}
 #else
     if (ioctl(0, TIOCGETP, &ttyb) == 0) {
-	(void) strcat(term, "/");
-	(void) strcat(term, speeds[ttyb.sg_ospeed]);
+	(void) strncat(term, "/", sizeof(term) - 1 - strlen(term));
+	(void) strncat(term, speeds[ttyb.sg_ospeed], sizeof(term) - 1 - strlen(term));
     }
 #endif
     (void) get_window_size(0, &winsize);
@@ -584,7 +637,7 @@ main(argc, argv)
     if (Fflag)
       authopts |= OPTS_FORWARDABLE_CREDS;
 
-    status = kcmd(&sock, &host, debug_port,
+    status = kcmd(&sock, &host, port,
 		  null_local_username ? "" : pwd->pw_name,
 		  name ? name : pwd->pw_name, term,
 		  0, "host", krb_realm,
@@ -592,28 +645,49 @@ main(argc, argv)
 		  0,		/* No need for sequence number */
 		  0,		/* No need for server seq # */
 		  &local, &foreign,
-		  authopts,
-		  0);		/* Not any port # */
+		  &auth_context, authopts,
+		  0,		/* Not any port # */
+		  0,
+		  &kcmd_proto);
     if (status) {
-	/* should check for KDC_PR_UNKNOWN, NO_TKT_FILE here -- XXX */
-	 if (status != -1) 
-	      fprintf(stderr,
-		      "%s: kcmd to host %s failed - %s\n",orig_argv[0], host,
-		      error_message(status));
-	 try_normal(orig_argv);
+	if (kcmd_proto == KCMD_NEW_PROTOCOL && encrypt_flag)
+	    /* Don't fall back to something less secure.  */
+	    exit (1);
+#ifdef KRB5_KRB4_COMPAT
+	fprintf(stderr, "Trying krb4 rlogin...\n");
+	status = k4cmd(&sock, &host, port,
+		       null_local_username ? "" : pwd->pw_name,
+		       name ? name : pwd->pw_name, term,
+		       0, &v4_ticket, "rcmd", krb_realm,
+		       &v4_cred, v4_schedule, &v4_msg_data, &local, &foreign,
+		       (encrypt_flag) ? KOPT_DO_MUTUAL : 0L, 0);
+	if (status)
+	    try_normal(orig_argv);
+	rcmd_stream_init_krb4(v4_cred.session, encrypt_flag, 1, 1);
+#else
+	try_normal(orig_argv);
+#endif
+    } else {
+	krb5_keyblock *key = 0;
+
+	if (kcmd_proto == KCMD_NEW_PROTOCOL) {
+	    do_inband = 1;
+
+	    status = krb5_auth_con_getlocalsubkey (bsd_context, auth_context,
+						   &key);
+	    if ((status || !key) && encrypt_flag)
+		try_normal(orig_argv);
+	}
+	if (key == 0)
+	    key = &cred->keyblock;
+
+	rcmd_stream_init_krb5(key, encrypt_flag, 1, 1, kcmd_proto);
     }
+	
     rem = sock;
     
-    /* setup eblock for des_read and write */
-    krb5_use_enctype(bsd_context, &eblock,cred->keyblock.enctype);
-    if ( status = krb5_process_key(bsd_context, &eblock,&cred->keyblock)) {
-	fprintf(stderr,
-		"%s: Cannot process session key : %s.\n",
-		orig_argv[0], error_message(status));
-	exit(1);
-    }
 #else
-    rem = rcmd(&host, debug_port,
+    rem = rcmd(&host, port,
 	       null_local_username ? "" : pwd->pw_name,
 	       name ? name : pwd->pw_name, term, 0);
 #endif /* KERBEROS */
@@ -980,7 +1054,7 @@ int signo;
  */
 writer()
 {
-    char c;
+    unsigned char c;
     register n;
     register bol = 1;               /* beginning of line */
     register local = 0;
@@ -1067,9 +1141,9 @@ writer()
 #endif
 
 	    if (c != cmdchar)
-	      (void) des_write(rem, &cmdchar, 1);
+	      (void) rcmd_stream_write(rem, &cmdchar, 1, 0);
 	}
-	if (des_write(rem, &c, 1) == 0) {
+	if (rcmd_stream_write(rem, &c, 1, 0) == 0) {
 	    prf("line gone");
 	    break;
 	}
@@ -1184,7 +1258,7 @@ sendwindow()
     wp->ws_col = htons(winsize.ws_col);
     wp->ws_xpixel = htons(winsize.ws_xpixel);
     wp->ws_ypixel = htons(winsize.ws_ypixel);
-    (void) des_write(rem, obuf, sizeof(obuf));
+    (void) rcmd_stream_write(rem, obuf, sizeof(obuf), 0);
 }
 
 
@@ -1200,15 +1274,16 @@ int	rcvcnt;
 int	rcvstate;
 int	ppid;
 
+/* returns 1 if flush, 0 otherwise */
 
-void oob()
+int server_message(mark)
+     int mark;
 {
 #ifndef POSIX_TERMIOS
     int out = FWRITE;
 #endif
-    int atmark, n;
+    int n;
     int rcvd = 0;
-    char waste[RLOGIN_BUFSIZ], mark;
 #ifdef POSIX_TERMIOS
     struct termios tty;
 #else
@@ -1218,9 +1293,7 @@ void oob()
     struct sgttyb sb;
 #endif
 #endif
-    mark = 0;
-    
-     recv(rem, &mark, 1, MSG_OOB);
+
     if (mark & TIOCPKT_WINDOW) {
 	/*
 	 * Let server know about window size changes
@@ -1278,27 +1351,62 @@ void oob()
 	(void) ioctl(1, TCFLSH, 1);
 #endif
 #endif
-	for (;;) {
-	    if (ioctl(rem, SIOCATMARK, &atmark) < 0) {
-		perror("ioctl");
-		break;
-	    }
-	    if (atmark)
-	      break;
-	    n = read(rem, waste, sizeof (waste));
-	    if (n <= 0)
-	      break;
-return;
-	}
+	return(1);
     }
-    
-    
+
+    return(0);
 }
 
+void oob()
+{
+    char mark;
+    char waste[RLOGIN_BUFSIZ];
+    int atmark;
 
+    mark = 0;
+    
+    recv(rem, &mark, 1, MSG_OOB);
+
+    if (server_message(mark)) {
+	if (ioctl(rem, SIOCATMARK, &atmark) < 0) {
+	    perror("ioctl");
+	    return;
+	}
+	if (!atmark)
+	    read(rem, waste, sizeof (waste));
+    }
+}
+
+/* two control messages are defined:
+
+   a double flag byte of 'o' indicates a one-byte message which is
+   identical to what was once carried out of band.  
+
+   a double flag byte of 'q' indicates a zero-byte message.  This
+   message is interpreted as two \377 data bytes.  This is just a
+   quote rule so that binary data from the server does not confuse the
+   client.  */
+
+int control(cp, n)
+     unsigned char *cp;
+     int n;
+{
+    if ((n >= 5) && (cp[2] == 'o') && (cp[3] == 'o')) {
+	if (server_message(cp[4]))
+	    return(-5);
+	return(5);
+    } else if ((n >= 4) && (cp[2] == 'q') && (cp[3] == 'q')) {
+	/* this is somewhat of a hack */
+	cp[2] = '\377';
+	cp[3] = '\377';
+	return(2);
+    }
+
+    return(0);
+}
 
 /*
- * reader: read from remote: line -> 1
+ * reader: read from remote: line -> 1 
  */
 reader(oldmask)
 #ifdef POSIX_SIGNALS
@@ -1313,8 +1421,9 @@ reader(oldmask)
     int pid = -getpid();
 #endif
 fd_set readset, excset, writeset;
-    int n, remaining;
+    int n, remaining, left;
     char *bufp = rcvbuf;
+    char *cp;
 
 #ifdef POSIX_SIGNALS
     struct sigaction sa;
@@ -1329,7 +1438,7 @@ fd_set readset, excset, writeset;
 #endif
     
     ppid = getppid();
-FD_ZERO(&readset);
+    FD_ZERO(&readset);
     FD_ZERO(&excset);
     FD_ZERO(&writeset);
 #ifdef POSIX_SIGNALS
@@ -1341,42 +1450,66 @@ FD_ZERO(&readset);
 #endif /* POSIX_SIGNALS */
 
     for (;;) {
-	if ((remaining = rcvcnt - (bufp - rcvbuf)) > 0)
-	{
+	if ((remaining = rcvcnt - (bufp - rcvbuf)) > 0) {
 	    FD_SET(1,&writeset);
 	    rcvstate = WRITING;
 	    FD_CLR(rem, &readset);
+	} else {
+	    bufp = rcvbuf;
+	    rcvcnt = 0;
+	    rcvstate = READING;
+	    FD_SET(rem,&readset);
+	    FD_CLR(1,&writeset);
 	}
-	else {
-	    
-	bufp = rcvbuf;
-	rcvcnt = 0;
-	 rcvstate = READING;
-FD_SET(rem,&readset);
-	FD_CLR(1,&writeset);
-	}
-	FD_SET(rem,&excset);
+	if (!do_inband)
+	    FD_SET(rem,&excset);
 	if (select(rem+1, &readset, &writeset, &excset, 0) > 0 ) {
-	    if (FD_ISSET(rem, &excset))
-  oob();
+	    if (!do_inband)
+		if (FD_ISSET(rem, &excset))
+		    oob();
 	    if (FD_ISSET(1,&writeset)) {
-    	    n = write(1, bufp, remaining);
-	    if (n < 0) {
-		if (errno != EINTR)
-		  return (-1);
-		continue;
+		n = write(1, bufp, remaining);
+		if (n < 0) {
+		    if (errno != EINTR)
+			return (-1);
+		    continue;
+		}
+		bufp += n;
+	    }
+	    if (FD_ISSET(rem, &readset)) {
+	  	rcvcnt = rcmd_stream_read(rem, rcvbuf, sizeof (rcvbuf), 0);
+		if (rcvcnt == 0)
+		    return (0);
+		if (rcvcnt < 0)
+		    goto error;
+
+		if (do_inband) {
+		    for (cp = rcvbuf; cp < rcvbuf+rcvcnt-1; cp++) {
+			if (cp[0] == '\377' &&
+			    cp[1] == '\377') {
+			    left = (rcvbuf+rcvcnt) - cp;
+			    n = control(cp, left);
+			    if (n < 0) {
+				left -= (-n);
+				rcvcnt = 0;
+				/* flush before, and (-n) bytes */
+				if (left > 0)
+				    memmove(rcvbuf, cp+(-n), left);
+				cp = rcvbuf-1;
+			    } else if (n) {
+				left -= n;
+				rcvcnt -= n;
+				if (left > 0)
+				    memmove(cp, cp+n, left);
+				cp--;
+			    }
+			}
+		    }
+		}
 	    }
-	    bufp += n;
-}
-if (FD_ISSET(rem, &readset)) {
-	  	rcvcnt = des_read(rem, rcvbuf, sizeof (rcvbuf));
-	if (rcvcnt == 0)
-	  return (0);
-	if (rcvcnt < 0)
-	  goto error;
-}
 	} else
-		  error: {
+error:
+	{
 	    if (errno == EINTR)
 	      continue;
 	    perror("read");
@@ -1390,67 +1523,70 @@ if (FD_ISSET(rem, &readset)) {
 mode(f)
 {
 #ifdef POSIX_TERMIOS
-	struct termios newtty;
+    struct termios newtty;
+#ifndef IEXTEN
+#define IEXTEN 0 /* No effect*/
+#endif
+#ifndef _POSIX_VDISABLE
+#define _POSIX_VDISABLE 0 /*A good guess at the disable-this-character character*/
+#endif
 
-	switch(f) {
-	case 0:
+    switch(f) {
+    case 0:
 #ifdef TIOCGLTC
-#ifndef solaris20
-		(void) ioctl(0, TIOCSLTC, (char *)&defltc);
+#if !defined(sun)
+	(void) ioctl(0, TIOCSLTC, (char *)&defltc);
 #endif
 #endif
-		(void) tcsetattr(0, TCSADRAIN, &deftty);
-		break;
-	case 1:
-		(void) tcgetattr(0, &newtty);
-/* was __svr4__ */
+	(void) tcsetattr(0, TCSADRAIN, &deftty);
+	break;
+    case 1:
+	(void) tcgetattr(0, &newtty);
+	/* was __svr4__ */
 #ifdef VLNEXT
 	/* there's a POSIX way of doing this, but do we need it general? */
-		newtty.c_cc[VLNEXT] = 0;
+	newtty.c_cc[VLNEXT] = _POSIX_VDISABLE;
 #endif
 		
-		newtty.c_lflag &= ~(ICANON|ISIG|ECHO);
-		if (!flow)
-		{
-			newtty.c_lflag &= ~(ICANON|ISIG|ECHO);
-			newtty.c_iflag &= ~(BRKINT|INLCR|ICRNL|ISTRIP);
-			/* newtty.c_iflag |=  (IXON|IXANY); */
-			newtty.c_iflag &= ~(IXON|IXANY);
-			newtty.c_oflag &= ~(OPOST);
-		} else {
-			newtty.c_lflag &= ~(ICANON|ISIG|ECHO);
-			newtty.c_iflag &= ~(INLCR|ICRNL);
-			/* newtty.c_iflag |=  (BRKINT|ISTRIP|IXON|IXANY); */
-			newtty.c_iflag &= ~(IXON|IXANY);
-			newtty.c_iflag |=  (BRKINT|ISTRIP);
-			newtty.c_oflag &= ~(ONLCR|ONOCR);
-			newtty.c_oflag |=  (OPOST);
-		}
-#ifdef TABDLY
-		/* preserve tab delays, but turn off XTABS */
-		if ((newtty.c_oflag & TABDLY) == TAB3)
-			newtty.c_oflag &= ~TABDLY;
-#endif
+	newtty.c_lflag &= ~(ICANON|ISIG|ECHO|IEXTEN);
+	newtty.c_iflag &= ~(ISTRIP|INLCR|ICRNL);
 
-		if (litout)
-			newtty.c_oflag &= ~OPOST;
-
-		newtty.c_cc[VMIN] = 1;
-		newtty.c_cc[VTIME] = 0;
-		(void) tcsetattr(0, TCSADRAIN, &newtty);
-#ifdef TIOCGLTC
-		/* Do this after the tcsetattr() in case this version
-		 * of termio supports the VSUSP or VDSUSP characters */
-#ifndef solaris20
-		/* this forces ICRNL under Solaris... */
-		(void) ioctl(0, TIOCSLTC, (char *)&noltc);
-#endif
-#endif
-		break;
-	default:
-		return;
-		/* NOTREACHED */
+	if (!flow) {
+	    newtty.c_iflag &= ~(BRKINT|IXON|IXANY);
+	    newtty.c_oflag &= ~(OPOST);
+	} else {
+	    /* XXX - should we set ixon ? */
+	    newtty.c_iflag &= ~(IXON|IXANY);
+	    newtty.c_iflag |=  (BRKINT);
+	    newtty.c_oflag &= ~(ONLCR|ONOCR);
+	    newtty.c_oflag |=  (OPOST);
 	}
+#ifdef TABDLY
+	/* preserve tab delays, but turn off XTABS */
+	if ((newtty.c_oflag & TABDLY) == TAB3)
+	    newtty.c_oflag &= ~TABDLY;
+#endif
+	if (!eight)
+	    newtty.c_iflag |= ISTRIP;
+	if (litout)
+	    newtty.c_oflag &= ~OPOST;
+
+	newtty.c_cc[VMIN] = 1;
+	newtty.c_cc[VTIME] = 0;
+	(void) tcsetattr(0, TCSADRAIN, &newtty);
+#ifdef TIOCGLTC
+	/* Do this after the tcsetattr() in case this version
+	 * of termio supports the VSUSP or VDSUSP characters */
+#if !defined(sun)
+	/* this forces ICRNL under Solaris... */
+	(void) ioctl(0, TIOCSLTC, (char *)&noltc);
+#endif
+#endif
+	break;
+    default:
+	return;
+	/* NOTREACHED */
+    }
 #else
     struct ltchars *ltc;
 #ifdef USE_TERMIO
@@ -1465,17 +1601,17 @@ mode(f)
     (void) ioctl(0, TIOCGETP, (char *)&sb);
     switch (f) {
 	
-      case 0:
+    case 0:
 #ifdef USE_TERMIO
 	/*
-	 **      remember whether IXON was set, so it can be restored
-	 **      when mode(1) is next done
-	 */
+	**      remember whether IXON was set, so it can be restored
+	**      when mode(1) is next done
+	*/
 	(void) ioctl(fileno(stdin), TIOCGETP, &ixon_state);
 	/*
-	 **      copy the initial modes we saved into sb; this is
-	 **      for restoring to the initial state
-	 */
+	**      copy the initial modes we saved into sb; this is
+	**      for restoring to the initial state
+	*/
 	(void)memcpy(&sb, &defmodes, sizeof(defmodes));
 	
 #else
@@ -1489,30 +1625,30 @@ mode(f)
 	ltc = &defltc;
 	break;
 	
-      case 1:
+    case 1:
 #ifdef USE_TERMIO
 	/*
-	 **      turn off output mappings
-	 */
+	**      turn off output mappings
+	*/
 	sb.c_oflag &= ~(ONLCR|OCRNL);
 	/*
-	 **      turn off canonical processing and character echo;
-	 **      also turn off signal checking -- ICANON might be
-	 **      enough to do this, but we're being careful
-	 */
+	**      turn off canonical processing and character echo;
+	**      also turn off signal checking -- ICANON might be
+	**      enough to do this, but we're being careful
+	*/
 	sb.c_lflag &= ~(ECHO|ICANON|ISIG);
 	sb.c_cc[VTIME] = 1;
 	sb.c_cc[VMIN] = 1;
 	if (eight)
-	  sb.c_iflag &= ~(ISTRIP);
+	    sb.c_iflag &= ~(ISTRIP);
 #ifdef TABDLY
 	/* preserve tab delays, but turn off tab-to-space expansion */
 	if ((sb.c_oflag & TABDLY) == TAB3)
-	  sb.c_oflag &= ~TAB3;
+	    sb.c_oflag &= ~TAB3;
 #endif
 	/*
-	 **  restore current flow control state
-	 */
+	**  restore current flow control state
+	*/
 	if ((ixon_state.c_iflag & IXON) && flow ) {
 	    sb.c_iflag |= IXON;
 	} else {
@@ -1523,15 +1659,15 @@ mode(f)
 	sb.sg_flags |= (!flow ? RAW : CBREAK);
 	/* preserve tab delays, but turn off XTABS */
 	if ((sb.sg_flags & TBDELAY) == XTABS)
-	  sb.sg_flags &= ~TBDELAY;
+	    sb.sg_flags &= ~TBDELAY;
 	sb.sg_kill = sb.sg_erase = -1;
 #ifdef LLITOUT
 	if (litout)
-	  lflags |= LLITOUT;
+	    lflags |= LLITOUT;
 #endif
 #ifdef LPASS8
 	if (eight)
-	  lflags |= LPASS8;
+	    lflags |= LPASS8;
 #endif /* LPASS8 */
 	tc = &notc;
 	sb.sg_flags &= ~defflags;
@@ -1540,7 +1676,7 @@ mode(f)
 	ltc = &noltc;
 	break;
 	
-      default:
+    default:
 	return;
     }
     (void) ioctl(0, TIOCSLTC, (char *)ltc);
@@ -1570,6 +1706,10 @@ void try_normal(argv)
      char **argv;
 {
     register char *host;
+#ifdef POSIX_SIGNALS
+    struct sigaction sa;
+    sigset_t mask;
+#endif
     
 #ifndef KRB5_ATHENA_COMPAT
     if (encrypt_flag)
@@ -1587,311 +1727,16 @@ void try_normal(argv)
     if (!strcmp(host, "rlogin"))
       argv++;
     
+#ifdef POSIX_SIGNALS
+    sigemptyset(&mask);
+    sigprocmask(SIG_SETMASK, &mask, NULL);
+#endif
+
     execv(UCB_RLOGIN, argv);
     perror("exec");
     exit(1);
 }
-
-
-
-char storage[2*RLOGIN_BUFSIZ];		/* storage for the decryption */
-int nstored = 0;
-char *store_ptr = storage;
-
-#ifndef OLD_VERSION
-
-int des_read(fd, buf, len)
-     int fd;
-     register char *buf;
-     int len;
-{
-    int nreturned = 0;
-    long net_len,rd_len;
-    int cc;
-    unsigned char len_buf[4];
-    
-    if (!encrypt_flag)
-      return(read(fd, buf, len));
-    
-    if (nstored >= len) {
-	memcpy(buf, store_ptr, len);
-	store_ptr += len;
-	nstored -= len;
-	return(len);
-    } else if (nstored) {
-	memcpy(buf, store_ptr, nstored);
-	nreturned += nstored;
-	buf += nstored;
-	len -= nstored;
-	nstored = 0;
-    }
-    
-#if 0
-    if ((cc = krb5_net_read(bsd_context, fd, (char *)&len_buf, 4)) != 4) {
-	/* XXX can't read enough, pipe must have closed */
-	return(0);
-    }
-    rd_len =
-	((len_buf[0]<<24) | (len_buf[1]<<16) | (len_buf[2]<<8) | len_buf[3]);
-#else
-	{
-	    unsigned char c;
-	    int gotzero = 0;
-
-	    /* See the comment in v4_des_read. */
-	    do {
-		cc = krb5_net_read(bsd_context, fd, &c, 1);
-		/* we should check for non-blocking here, but we'd have
-		   to make it save partial reads as well. */
-		if (cc <= 0) return cc; /* read error */
-		if (cc == 1) {
-		    if (c == 0) gotzero = 1;
-		}
-	    } while (!gotzero);
-
-	    if ((cc = krb5_net_read(bsd_context, fd, &c, 1)) != 1) return 0;
-	    rd_len = c;
-	    if ((cc = krb5_net_read(bsd_context, fd, &c, 1)) != 1) return 0;
-	    rd_len = (rd_len << 8) | c;
-	    if ((cc = krb5_net_read(bsd_context, fd, &c, 1)) != 1) return 0;
-	    rd_len = (rd_len << 8) | c;
-	}
-
 #endif
-    net_len = krb5_encrypt_size(rd_len,eblock.crypto_entry);
-    if ((net_len <= 0) || (net_len > sizeof(des_inbuf))) {
-	/* preposterous length; assume out-of-sync; only
-	   recourse is to close connection, so return 0 */
-	fprintf(stderr,"Read size problem.\n");
-	return(0);
-    }
-    if ((cc = krb5_net_read(bsd_context, fd, desinbuf.data, net_len)) != net_len) {
-	/* pipe must have closed, return 0 */
-	fprintf(stderr,
-		"Read error: length received %d != expected %d.\n",
-		cc,net_len);
-	return(0);
-    }
-    /* decrypt info */
-    if ((krb5_decrypt(bsd_context, desinbuf.data,
-		      (krb5_pointer) storage,
-		      net_len,
-		      &eblock, 0))) {
-	fprintf(stderr,"Cannot decrypt data from network.\n");
-	return(0);
-    }
-    store_ptr = storage;
-    nstored = rd_len;
-    if (nstored > len) {
-	memcpy(buf, store_ptr, len);
-	nreturned += len;
-	store_ptr += len;
-	nstored -= len;
-    } else {
-	memcpy(buf, store_ptr, nstored);
-	nreturned += nstored;
-	nstored = 0;
-    }
-    
-    return(nreturned);
-}
-
-
-
-int des_write(fd, buf, len)
-     int fd;
-     char *buf;
-     int len;
-{
-  unsigned char *len_buf = (unsigned char *) des_outpkt;
-    
-  if (!encrypt_flag)
-      return(write(fd, buf, len));
-    
-    
-    desoutbuf.length = krb5_encrypt_size(len,eblock.crypto_entry);
-    if (desoutbuf.length > sizeof(des_outpkt)-4){
-      	fprintf(stderr,"Write size problem.\n");
-	return(-1);
-    }
-    if ((krb5_encrypt(bsd_context, (krb5_pointer)buf,
-		      desoutbuf.data,
-		      len,
-		      &eblock,
-		      0))){
-      	fprintf(stderr,"Write encrypt problem.\n");
-	return(-1);
-    }
-
-    len_buf[0] = (len & 0xff000000) >> 24;
-    len_buf[1] = (len & 0xff0000) >> 16;
-    len_buf[2] = (len & 0xff00) >> 8;
-    len_buf[3] = (len & 0xff);
-
-    if (write(fd, des_outpkt,desoutbuf.length+4) != desoutbuf.length+4){
-      fprintf(stderr,"Could not write out all data\n");
-	return(-1);
-    }
-    else return(len);
-}
-
-
-
-#else /* Original version  placed here so that testing could be done
-	 to determine why rlogin with encryption on is slower with
-	 version 5 as compared to version 4. */
-
-#define ENCRYPT 1
-#define DECRYPT 0
-
-
-
-int des_read(fd, buf, len)
-     int fd;
-     register char *buf;
-     int len;
-{
-    int nreturned = 0;
-    long net_len, rd_len;
-    int cc;
-    unsigned char len_buf[4];
-    
-    if (!encrypt_flag)
-      return(read(fd, buf, len));
-    
-    if (nstored >= len) {
-	memcpy(buf, store_ptr, len);
-	store_ptr += len;
-	nstored -= len;
-	return(len);
-    } else if (nstored) {
-	memcpy(buf, store_ptr, nstored);
-	nreturned += nstored;
-	buf += nstored;
-	len -= nstored;
-	nstored = 0;
-    }
-#if 0
-    if ((cc = krb5_net_read(bsd_context, fd, len_buf, 4)) != 4) {
-	/* XXX can't read enough, pipe must have closed */
-	return(0);
-    }
-    net_len =
-	((len_buf[0]<<24) | (len_buf[1]<<16) | (len_buf[2]<<8) | len_buf[3]);
-#else
-	{
-	    unsigned char c;
-	    int gotzero = 0;
-
-	    /* See the comment in v4_des_read. */
-	    do {
-		cc = krb5_net_read(bsd_context, fd, &c, 1);
-		/* we should check for non-blocking here, but we'd have
-		   to make it save partial reads as well. */
-		if (cc < 0) return 0; /* read error */
-		if (cc == 1) {
-		    if (c == 0) gotzero = 1;
-		}
-	    } while (!gotzero);
-
-	    if ((cc = krb5_net_read(bsd_context, fd, &c, 1)) != 1) return 0;
-	    net_len = c;
-	    if ((cc = krb5_net_read(bsd_context, fd, &c, 1)) != 1) return 0;
-	    net_len = (net_len << 8) | c;
-	    if ((cc = krb5_net_read(bsd_context, fd, &c, 1)) != 1) return 0;
-	    net_len = (net_len << 8) | c;
-	}
-#endif
-    if (net_len < 0 || net_len > sizeof(des_inbuf)) {
-	/* XXX preposterous length, probably out of sync.
-	   act as if pipe closed */
-	return(0);
-    }
-    /* the writer tells us how much real data we are getting, but
-       we need to read the pad bytes (8-byte boundary) */
-#ifdef NOROUNDUP
-    rd_len = ((((net_len)+((8)-1))/(8))*(8));
-#else
-    rd_len = roundup(net_len, 8);
-#endif
-    if ((cc = krb5_net_read(bsd_context, fd, des_inbuf, rd_len)) != rd_len) {
-	/* pipe must have closed, return 0 */
-	return(0);
-    }
-    (void) mit_des_cbc_encrypt(
-			       des_inbuf,
-			       storage,
-			       (net_len < 8) ? 8 : net_len,
-			       eblock.priv,
-			       eblock.key->contents,
-			       DECRYPT);
-    /*
-     * when the cleartext block is < 8 bytes, it is "right-justified"
-     * in the block, so we need to adjust the pointer to the data
-     */
-    if (net_len < 8)
-      store_ptr = storage + 8 - net_len;
-    else
-      store_ptr = storage;
-    nstored = net_len;
-    if (nstored > len) {
-	memcpy(buf, store_ptr, len);
-	nreturned += len;
-	store_ptr += len;
-	nstored -= len;
-    } else {
-	memcpy(buf, store_ptr, nstored);
-	nreturned += nstored;
-	nstored = 0;
-    }
-    return(nreturned);
-}
-
-
-
-int des_write(fd, buf, len)
-     int fd;
-     char *buf;
-     int len;
-{
-    static char garbage_buf[8];
-    unsigned char len_buf[4];
-    
-    if (!encrypt_flag)
-      return(write(fd, buf, len));
-    
-#define min(a,b) ((a < b) ? a : b)
-    
-    if (len < 8) {
-	krb5_random_confounder(bsd_context, 8 - len, &garbage_buf);
-	/* this "right-justifies" the data in the buffer */
-	(void) memcpy(garbage_buf + 8 - len, buf, len);
-    }
-    
-    (void) mit_des_cbc_encrypt((len < 8) ? garbage_buf : buf,
-			       des_outbuf,
-			       (len < 8) ? 8 : len,
-			       eblock.priv,
-			       eblock.key->contents,
-			       ENCRYPT);
-    
-    /* tell the other end the real amount, but send an 8-byte padded
-       packet */
-    len_buf[0] = (len & 0xff000000) >> 24;
-    len_buf[1] = (len & 0xff0000) >> 16;
-    len_buf[2] = (len & 0xff00) >> 8;
-    len_buf[3] = (len & 0xff);
-    (void) write(fd, len_buf, 4);
-#ifdef NOROUNDUP
-    (void) write(fd, des_outbuf, ((((len)+((8)-1))/(8))*(8)));
-#else
-    (void) write(fd, des_outbuf, roundup(len,8));
-#endif
-    return(len);
-}
-
-#endif /* OLD_VERSION */
-#endif /* KERBEROS */
 
 
 
