@@ -19,6 +19,32 @@
  * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
  */
 
+/*
+ * Copyright (C) 1998 by the FundsXpress, INC.
+ * 
+ * All rights reserved.
+ * 
+ * Export of this software from the United States of America may require
+ * a specific license from the United States Government.  It is the
+ * responsibility of any person or organization contemplating export to
+ * obtain such a license before exporting.
+ * 
+ * WITHIN THAT CONSTRAINT, permission to use, copy, modify, and
+ * distribute this software and its documentation for any purpose and
+ * without fee is hereby granted, provided that the above copyright
+ * notice appear in all copies and that both that copyright notice and
+ * this permission notice appear in supporting documentation, and that
+ * the name of FundsXpress. not be used in advertising or publicity pertaining
+ * to distribution of the software without specific, written prior
+ * permission.  FundsXpress makes no representations about the suitability of
+ * this software for any purpose.  It is provided "as is" without express
+ * or implied warranty.
+ * 
+ * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
+ * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
+ * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
+ */
+
 #ifndef lint
 char copyright[] =
   "@(#) Copyright (c) 1983 The Regents of the University of California.\n\
@@ -105,7 +131,10 @@ char copyright[] =
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#ifndef KERBEROS
+/* Ultrix doesn't protect it vs multiple inclusion, and krb.h includes it */
 #include <sys/socket.h>
+#endif
 #include <sys/ioctl.h>
 #include <sys/wait.h>
 #include <sys/file.h>
@@ -146,8 +175,11 @@ char copyright[] =
 #else
 #include <sgtty.h>
 #endif
-     
+
+#ifndef KERBEROS
+/* Ultrix doesn't protect it vs multiple inclusion, and krb.h includes it */
 #include <netdb.h>
+#endif
 #include <syslog.h>
 #include <string.h>
 #include <sys/param.h>
@@ -199,7 +231,7 @@ struct winsize {
 
 #ifdef KERBEROS
      
-#include "krb5.h"
+#include <krb5.h>
 #include <kerberosIV/krb.h>
 #include <libpty.h>
 #ifdef HAVE_UTMP_H
@@ -216,28 +248,11 @@ int non_privileged = 0; /* set when connection is seen to be from */
 
 AUTH_DAT	*v4_kdata;
 Key_schedule v4_schedule;
-int v4_des_read(), v4_des_write();
-
-#define RLOGIND_BUFSIZ 5120
-
-int v5_des_read(), v5_des_write();
 
 #include "com_err.h"
+#include "defines.h"
      
 #define SECURE_MESSAGE  "This rlogin session is using DES encryption for all data transmissions.\r\n"
-/*
- * Note that the encrypted rlogin packets take the form of a four-byte
- *length followed by encrypted data.  On writing the data out, a significant
- * performance penalty is suffered (at least one RTT per character, two if we
- * are waiting for a shell to echo) by writing the data separately from the 
- * length.  So, unlike the input buffer, which just contains the output
- * data, the output buffer represents the entire packet.
- */
- int (*des_read)(), (*des_write)();
-char des_inbuf[2*RLOGIND_BUFSIZ]; /* needs to be > largest read size */
-char des_outpkt[2*RLOGIND_BUFSIZ+4];/* needs to be > largest write size */
-krb5_data desinbuf,desoutbuf;
-krb5_encrypt_block eblock;        /* eblock for encrypt/decrypt */
 
 krb5_authenticator      *kdata;
 krb5_ticket     *ticket = 0;
@@ -246,11 +261,9 @@ krb5_ccache ccache = NULL;
 
 krb5_keytab keytab = NULL;
 
-#define ARGSTR	"k54ciepPD:S:M:L:w:?"
+#define ARGSTR	"k54ciepPD:S:M:L:fw:?"
 #else /* !KERBEROS */
-#define ARGSTR	"rpPD:?"
-#define (*des_read)  read
-#define (*des_write) write
+#define ARGSTR	"rpPD:f?"
 #endif /* KERBEROS */
 
 #ifndef LOGIN_PROGRAM
@@ -289,6 +302,7 @@ char		term[64];
 char            rhost_name[MAXDNAME];
 char		rhost_addra[16];
 krb5_principal  client;
+int		do_inband = 0;
 
 int	reapchild();
 char 	*progname;
@@ -338,6 +352,7 @@ int main(argc, argv)
     char *options;
     int debug_port = 0;
     int fd;
+    int do_fork = 0;
 #ifdef KERBEROS
     krb5_error_code status;
 #endif
@@ -367,7 +382,7 @@ int main(argc, argv)
     
     /* Analyse parameters. */
     opterr = 0;
-    while ((ch = getopt(argc, argv, ARGSTR)) != EOF)
+    while ((ch = getopt(argc, argv, ARGSTR)) != -1)
       switch (ch) {
 #ifdef KERBEROS
 	case 'k':
@@ -424,11 +439,8 @@ int main(argc, argv)
 	case 'L':
 	  login_program = optarg;
 	  break;
-        case 'u':
-	  maxhostlen = atoi(optarg);
-	  break;
-        case 'I':
-	  always_ip = 1;
+        case 'f':
+	  do_fork = 1;
 	  break;
 	case 'w':
 	  if (!strcmp(optarg, "ip"))
@@ -463,65 +475,97 @@ int main(argc, argv)
     
     fromlen = sizeof (from);
 
-     if (debug_port) {
-	 int s;
-	 struct sockaddr_in sin;
-	 
-	 if ((s = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) < 0) {
-	     fprintf(stderr, "Error in socket: %s\n", strerror(errno));
-	     exit(2);
-	 }
-	 
-	 memset((char *) &sin, 0,sizeof(sin));
-	 sin.sin_family = AF_INET;
-	 sin.sin_port = htons(debug_port);
-	 sin.sin_addr.s_addr = INADDR_ANY;
-	 
-	 (void) setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
-			   (char *)&on, sizeof(on));
+    if (debug_port || do_fork) {
+	int s;
+	struct servent *ent;
+	struct sockaddr_in sin;
 
-	 if ((bind(s, (struct sockaddr *) &sin, sizeof(sin))) < 0) {
-	     fprintf(stderr, "Error in bind: %s\n", strerror(errno));
-	     exit(2);
-	 }
-	 
-	 if ((listen(s, 5)) < 0) {
-	     fprintf(stderr, "Error in listen: %s\n", strerror(errno));
-	     exit(2);
-	 }
-	 
-	 if ((fd = accept(s, (struct sockaddr *) &from, &fromlen)) < 0) {
-	     fprintf(stderr, "Error in accept: %s\n", strerror(errno));
-	     exit(2);
-	 }
-	 
-	 close(s);
-     } 
-     else {
-	 if (getpeername(0, (struct sockaddr *)&from, &fromlen) < 0) {
-	     syslog(LOG_ERR,"Can't get peer name of remote host: %m");
+	if (!debug_port) {
+	    if (do_encrypt) {
+		ent = getservbyname("eklogin", "tcp");
+		if (ent == NULL)
+		    debug_port = 2105;
+		else
+		    debug_port = ent->s_port;
+	    } else {
+		ent = getservbyname("klogin", "tcp");
+		if (ent == NULL)
+		    debug_port = 543;
+		else
+		    debug_port = ent->s_port;
+	    }
+	}
+	if ((s = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) < 0) {
+	    fprintf(stderr, "Error in socket: %s\n", strerror(errno));
+	    exit(2);
+	}
+	memset((char *) &sin, 0,sizeof(sin));
+	sin.sin_family = AF_INET;
+	sin.sin_port = htons(debug_port);
+	sin.sin_addr.s_addr = INADDR_ANY;
+
+	if (!do_fork)
+	    (void) setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
+			      (char *)&on, sizeof(on));
+
+	if ((bind(s, (struct sockaddr *) &sin, sizeof(sin))) < 0) {
+	    fprintf(stderr, "Error in bind: %s\n", strerror(errno));
+	    exit(2);
+	}
+
+	if ((listen(s, 5)) < 0) {
+	    fprintf(stderr, "Error in listen: %s\n", strerror(errno));
+	    exit(2);
+	}
+
+	if (do_fork) {
+	    if (daemon(0, 0)) {
+		fprintf(stderr, "daemon() failed\n");
+		exit(2);
+	    }
+	    while (1) {
+		int child_pid;
+
+		fd = accept(s, (struct sockaddr *) &from, &fromlen);
+		if (s < 0) {
+		    if (errno != EINTR)
+			syslog(LOG_ERR, "accept: %s", error_message(errno));
+		    continue;
+		}
+		child_pid = fork();
+		switch (child_pid) {
+		case -1:
+		    syslog(LOG_ERR, "fork: %s", error_message(errno));
+		case 0:
+		    (void) close(s);
+		    doit(fd, &from);
+		    close(fd);
+		    exit(0);
+		default:
+		    wait(0);
+		    close(fd);
+		}
+	    }
+	}
+
+	if ((fd = accept(s, (struct sockaddr *) &from, &fromlen)) < 0) {
+	    fprintf(stderr, "Error in accept: %s\n", strerror(errno));
+	    exit(2);
+	}
+
+	close(s);
+    } else {			/* !do_fork && !debug_port */
+	if (getpeername(0, (struct sockaddr *)&from, &fromlen) < 0) {
+	    syslog(LOG_ERR,"Can't get peer name of remote host: %m");
 #ifdef STDERR_FILENO
-	     fatal(STDERR_FILENO, "Can't get peer name of remote host", 1);
+	    fatal(STDERR_FILENO, "Can't get peer name of remote host", 1);
 #else
-	     fatal(2, "Can't get peer name of remote host", 1);
+	    fatal(2, "Can't get peer name of remote host", 1);
 #endif
-	 }
-	 fd = 0;
-     }
-
-    if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE,
-		   (const char *) &on, sizeof (on)) < 0)
-      syslog(LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");
-    if (auth_ok == 0) {
-      syslog(LOG_CRIT, "No authentication systems were enabled; all connections will be refused.");
-      fatal(fd, "All authentication systems disabled; connection refused.");
+	}
+	fd = 0;
     }
 
-    if (checksum_required&&checksum_ignored) {
-      syslog( LOG_CRIT, "Checksums are required and ignored; these options are mutually exclusive--check the documentation.");
-      fatal(fd, "Configuration error: mutually exclusive options specified");
-    }
-    
     doit(fd, &from);
     return 0;
 }
@@ -560,6 +604,19 @@ void doit(f, fromp)
     int syncpipe[2];
 
     netf = -1;
+    if (setsockopt(f, SOL_SOCKET, SO_KEEPALIVE,
+		   (const char *) &on, sizeof (on)) < 0)
+	syslog(LOG_WARNING, "setsockopt (SO_KEEPALIVE): %m");
+    if (auth_ok == 0) {
+	syslog(LOG_CRIT, "No authentication systems were enabled; all connections will be refused.");
+	fatal(f, "All authentication systems disabled; connection refused.");
+    }
+
+    if (checksum_required&&checksum_ignored) {
+	syslog( LOG_CRIT, "Checksums are required and ignored; these options are mutually exclusive--check the documentation.");
+	fatal(f, "Configuration error: mutually exclusive options specified");
+    }
+    
     alarm(60);
     read(f, &c, 1);
     
@@ -594,13 +651,7 @@ void doit(f, fromp)
     if (fromp->sin_family != AF_INET)
       fatal(f, "Permission denied - Malformed from address\n");
     
-#ifdef KERBEROS
-
-	/* setup des buffers */
-    desinbuf.data = des_inbuf;
-    desoutbuf.data = des_outpkt+4;    /* Set up des buffers */
-
-#else /* !KERBEROS */
+#ifndef KERBEROS
     if (fromp->sin_port >= IPPORT_RESERVED ||
 	fromp->sin_port < IPPORT_RESERVED/2)
       fatal(f, "Permission denied - Connection from bad port");
@@ -617,6 +668,7 @@ void doit(f, fromp)
     getstr(f, rusername, sizeof(rusername), "remuser");
     getstr(f, lusername, sizeof(lusername), "locuser");
     getstr(f, term, sizeof(term), "Terminal type");
+    rcmd_stream_init_normal();
 #endif
     
     write(f, "", 1);
@@ -656,12 +708,12 @@ void doit(f, fromp)
 #if defined(POSIX_TERMIOS) && !defined(ultrix)
 	tcgetattr(t,&new_termio);
 #if !defined(USE_LOGIN_F)
-	new_termio.c_lflag &=  ~(ICANON|ECHO|ISIG|IEXTEN);
+	new_termio.c_lflag &= ~(ICANON|ECHO|ISIG|IEXTEN);
 	new_termio.c_iflag &= ~(IXON|IXANY|BRKINT|INLCR|ICRNL);
 #else
 	new_termio.c_lflag |= (ICANON|ECHO|ISIG|IEXTEN);
 	new_termio.c_oflag |= (ONLCR|OPOST);
-	new_termio.c_iflag|= (IXON|IXANY|BRKINT|INLCR|ICRNL);
+	new_termio.c_iflag |= (IXON|IXANY|BRKINT|INLCR|ICRNL);
 #endif /*Do we need binary stream?*/
 	new_termio.c_iflag &= ~(ISTRIP);
 	/* new_termio.c_iflag = 0; */
@@ -764,7 +816,7 @@ void doit(f, fromp)
 				    stripdomain, always_ip,
 				    &rhost_sane);
     if (retval)
-        fatalperror(2, "failed make_sane_hostname");
+        fatalperror(f, "failed make_sane_hostname");
     if (passwd_req)
         execl(login_program, "login", "-p", "-h", rhost_sane,
           lusername, 0);
@@ -774,8 +826,9 @@ void doit(f, fromp)
 #else /* USE_LOGIN_F */
 	execl(login_program, "login", "-r", rhost_sane, 0);
 #endif /* USE_LOGIN_F */
-	
-	fatalperror(2, login_program);
+	syslog(LOG_ERR, "failed exec of %s: %s",
+	       login_program, error_message(errno));
+	fatalperror(f, login_program);
 	/*NOTREACHED*/
     } /* if (pid == 0) */
 
@@ -799,7 +852,7 @@ void doit(f, fromp)
     
 #if defined(KERBEROS) 
     if (do_encrypt) {
-	if (((*des_write)(f, SECURE_MESSAGE, sizeof(SECURE_MESSAGE))) < 0){
+	if (rcmd_stream_write(f, SECURE_MESSAGE, sizeof(SECURE_MESSAGE), 0) < 0){
 	    sprintf(buferror, "Cannot encrypt-write network.");
 	    fatal(p,buferror);
 	}
@@ -824,7 +877,7 @@ void doit(f, fromp)
 #endif
 
     
-#if!defined(USE_LOGIN_F)
+#if !defined(USE_LOGIN_F)
     /* Pass down rusername and lusername to login. */
     (void) write(p, rusername, strlen(rusername) +1);
     (void) write(p, lusername, strlen(lusername) +1);
@@ -853,6 +906,31 @@ unsigned char	oobdata[] = {TIOCPKT_WINDOW};
 char    oobdata[] = {0};
 #endif
 
+int sendoob(fd, byte)
+     int fd;
+     char *byte;
+{
+    char message[5];
+    int cc;
+
+    if (do_inband) {
+	message[0] = '\377';
+	message[1] = '\377';
+	message[2] = 'o';
+	message[3] = 'o';
+	message[4] = *byte;
+
+	cc = rcmd_stream_write(fd, message, sizeof(message), 0);
+	while (cc < 0 && ((errno == EWOULDBLOCK) || (errno == EAGAIN))) {
+	    /* also shouldn't happen */
+	    sleep(5);
+	    cc = rcmd_stream_write(fd, message, sizeof(message), 0);
+	}
+    } else {
+	send(fd, byte, 1, MSG_OOB);
+    }
+}
+
 /*
  * Handle a "control" request (signaled by magic being present)
  * in the data stream.  For now, we are only willing to handle
@@ -896,7 +974,7 @@ int control(pty, cp, n)
 void protocol(f, p)
      int f, p;
 {
-    unsigned char pibuf[BUFSIZ], fibuf[BUFSIZ], *pbp, *fbp;
+    unsigned char pibuf[BUFSIZ], qpibuf[BUFSIZ*2], fibuf[BUFSIZ], *pbp, *fbp;
     register pcc = 0, fcc = 0;
     int cc;
     char cntl;
@@ -931,7 +1009,7 @@ void protocol(f, p)
     signal(SIGTTOU, SIG_IGN);
 #endif
 #ifdef TIOCSWINSZ
-    send(f, oobdata, 1, MSG_OOB);	/* indicate new rlogin */
+    sendoob(f, oobdata);
 #endif
     for (;;) {
 	fd_set ibits, obits, ebits;
@@ -949,7 +1027,6 @@ void protocol(f, p)
 		FD_SET(f, &obits);
 	    else
 		FD_SET(p, &ibits);
-	FD_SET(p, &ebits);
 	
 	if (select(8*sizeof(ibits), &ibits, &obits, &ebits, 0) < 0) {
 	    if (errno == EINTR)
@@ -957,43 +1034,32 @@ void protocol(f, p)
 	    fatalperror(f, "select");
 	}
 #define	pkcontrol(c)	((c)&(TIOCPKT_FLUSHWRITE|TIOCPKT_NOSTOP|TIOCPKT_DOSTOP))
-	if (FD_ISSET(p, &ebits)) {
-	    cc = read(p, &cntl, 1);
-	    if (cc == 1 && pkcontrol(cntl)) {
-		cntl |= oobdata[0];
-		send(f, &cntl, 1, MSG_OOB);
-		if (cntl & TIOCPKT_FLUSHWRITE) {
-		    pcc = 0;
-		    FD_CLR(p, &ibits);
-		}
-	    }
-	}
 	if (FD_ISSET(f, &ibits)) {
-	    fcc = (*des_read)(f, fibuf, sizeof (fibuf));
-	    if (fcc < 0 && ((errno == EWOULDBLOCK) || (errno == EAGAIN)))
-	      fcc = 0;
-	    else {
+	    fcc = rcmd_stream_read(f, fibuf, sizeof (fibuf), 0);
+	    if (fcc < 0 && ((errno == EWOULDBLOCK) || (errno == EAGAIN))) {
+		fcc = 0;
+	    } else {
 		register unsigned char *cp;
 		int left, n;
 		
 		if (fcc <= 0)
-		  break;
+		    break;
 		fbp = fibuf;
 		
-	      top:
-		for (cp = fibuf; cp < fibuf+fcc-1; cp++)
-		  if (cp[0] == magic[0] &&
-		      cp[1] == magic[1]) {
-		      left = fcc - (cp-fibuf);
-		      n = control(p, cp, left);
-		      if (n) {
-			  left -= n;
-			  if (left > 0)
-			    memmove(cp, cp+n, left);
-			  fcc -= n;
-			  goto top; /* n^2 */
-		      }
-		  }
+		for (cp = fibuf; cp < fibuf+fcc-1; cp++) {
+		    if (cp[0] == magic[0] &&
+			cp[1] == magic[1]) {
+			left = (fibuf+fcc) - cp;
+			n = control(p, cp, left);
+			if (n) {
+			    left -= n;
+			    fcc -= n;
+			    if (left > 0)
+				memmove(cp, cp+n, left);
+			    cp--;
+			}
+		    }
+		}
 	    }
 	}
 	
@@ -1008,26 +1074,56 @@ void protocol(f, p)
 	if (FD_ISSET(p, &ibits)) {
 	    pcc = read(p, pibuf, sizeof (pibuf));
 	    pbp = pibuf;
-	    if (pcc < 0 && ((errno == EWOULDBLOCK) || (errno == EAGAIN)))
-	      pcc = 0;
-	    else if (pcc <= 0)
-	      break;
+	    if (pcc < 0 && ((errno == EWOULDBLOCK) || (errno == EAGAIN))) {
+		pcc = 0;
+	    } else if (pcc <= 0) {
+		break;
+	    }
 #ifdef TIOCPKT
 	    else if (tiocpkt_on) {
-	      if (pibuf[0] == 0)
-	        pbp++, pcc--;
-	      else {
-		  if (pkcontrol(pibuf[0])) {
-		      pibuf[0] |= oobdata[0];
-		      send(f, &pibuf[0], 1, MSG_OOB);
-		  }
-		  pcc = 0;
-	      }
+		if (pibuf[0] == 0) {
+		    pbp++, pcc--;
+		} else {
+		    if (pkcontrol(pibuf[0])) {
+			pibuf[0] |= oobdata[0];
+			sendoob(f, pibuf);
+		    }
+		    pcc = 0;
+		}
 	    }
 #endif
+
+	    /* quote any double-\377's if necessary */
+
+	    if (do_inband) {
+		unsigned char *qpbp;
+		int qpcc, i;
+
+		qpbp = qpibuf;
+		qpcc = 0;
+
+		for (i=0; i<pcc;) {
+		    if (pbp[i] == 0377u && (i+1)<pcc && pbp[i+1] == 0377u) {
+			qpbp[qpcc] = '\377';
+			qpbp[qpcc+1] = '\377';
+			qpbp[qpcc+2] = 'q';
+			qpbp[qpcc+3] = 'q';
+			i += 2;
+			qpcc += 4;
+		    } else {
+			qpbp[qpcc] = pbp[i];
+			i++;
+			qpcc++;
+		    }
+		}
+
+		pbp = qpbp;
+		pcc = qpcc;
+	    }
 	}
+
 	if (FD_ISSET(f, &obits) && pcc > 0) {
-	    cc = (*des_write)(f, pbp, pcc);
+	    cc = rcmd_stream_write(f, pbp, pcc, 0);
 	    if (cc < 0 && ((errno == EWOULDBLOCK) || (errno == EAGAIN))) {
 		/* also shouldn't happen */
 		sleep(5);
@@ -1066,7 +1162,7 @@ void fatal(f, msg)
     buf[0] = '\01';		/* error indicator */
     (void) sprintf(buf + 1, "%s: %s.\r\n",progname, msg);
     if ((f == netf) && (pid > 0))
-      (void) (*des_write)(f, buf, strlen(buf));
+      (void) rcmd_stream_write(f, buf, strlen(buf), 0);
     else
       (void) write(f, buf, strlen(buf));
     syslog(LOG_ERR,"%s\n",msg);
@@ -1189,6 +1285,8 @@ do_krb_login(host_addr, hostname)
     /* NOTREACHED */
 }
 
+#endif /* KERBEROS */
+
 
 
 void getstr(fd, buf, cnt, err)
@@ -1214,174 +1312,14 @@ void getstr(fd, buf, cnt, err)
 
 
 
-char storage[2*RLOGIND_BUFSIZ];             /* storage for the decryption */
-int nstored = 0;
-char *store_ptr = storage;
-
-int
-v5_des_read(fd, buf, len)
-     int fd;
-     register char *buf;
-     int len;
-{
-    int nreturned = 0;
-    krb5_ui_4 net_len,rd_len;
-    int cc,retry;
-#if 0
-    unsigned char len_buf[4];
-#endif
-    
-    if (!do_encrypt)
-      return(read(fd, buf, len));
-    
-    if (nstored >= len) {
-	memcpy(buf, store_ptr, len);
-	store_ptr += len;
-	nstored -= len;
-	return(len);
-    } else if (nstored) {
-	memcpy(buf, store_ptr, nstored);
-	nreturned += nstored;
-	buf += nstored;
-	len -= nstored;
-	nstored = 0;
-    }
-    
-#if 0
-    if ((cc = krb5_net_read(bsd_context, fd, (char *)len_buf, 4)) != 4) {
-	if ((cc < 0)  && ((errno == EWOULDBLOCK) || (errno == EAGAIN)))
-	    return(cc);
-	/* XXX can't read enough, pipe must have closed */
-	return(0);
-    }
-    rd_len =
-	(((krb5_ui_4)len_buf[0]<<24) |
-	 ((krb5_ui_4)len_buf[1]<<16) |
-	 ((krb5_ui_4)len_buf[2]<<8) |
-	 (krb5_ui_4)len_buf[3]);
-#else
-	{
-	    unsigned char c;
-	    int gotzero = 0;
-
-	    /* See the comment in v4_des_read. */
-	    do {
-		cc = krb5_net_read(bsd_context, fd, &c, 1);
-		/* we should check for non-blocking here, but we'd have
-		   to make it save partial reads as well. */
-		if (cc <= 0) return 0; /* read error */
-		if (cc == 1) {
-		    if (c == 0) gotzero = 1;
-		}
-	    } while (!gotzero);
-
-	    if ((cc = krb5_net_read(bsd_context, fd, &c, 1)) != 1) return 0;
-	    rd_len = c;
-	    if ((cc = krb5_net_read(bsd_context, fd, &c, 1)) != 1) return 0;
-	    rd_len = (rd_len << 8) | c;
-	    if ((cc = krb5_net_read(bsd_context, fd, &c, 1)) != 1) return 0;
-	    rd_len = (rd_len << 8) | c;
-	}
-#endif
-    net_len = krb5_encrypt_size(rd_len,eblock.crypto_entry);
-    /* note net_len is unsigned */
-    if (net_len > sizeof(des_inbuf)) {
-	/* XXX preposterous length, probably out of sync.
-	   act as if pipe closed */
-	syslog(LOG_ERR,"Read size problem.");
-	return(0);
-    }
-    retry = 0;
-  datard:
-    if ((cc = krb5_net_read(bsd_context,fd,desinbuf.data,net_len)) != net_len) {
-	/* XXX can't read enough, pipe must have closed */
-	if ((cc < 0)  && ((errno == EWOULDBLOCK) || (errno == EAGAIN))) {
-	    retry++;
-	    sleep(1);
-	    if (retry > MAXRETRIES){
-		syslog(LOG_ERR,
-		       "des_read retry count exceeded %d\n",
-		       retry);
-		return(0);
-	    }
-	    goto datard;
-	}
-	syslog(LOG_ERR,
-	       "Read data received %d != expected %d.",
-	       cc, net_len);
-	return(0);
-    }
-    /* decrypt info */
-    if ((krb5_decrypt(bsd_context, desinbuf.data,
-		      (krb5_pointer) storage,
-		      net_len,
-		      &eblock, 0))) {
-	syslog(LOG_ERR,"Read decrypt problem.");
-	return(0);
-    }
-    store_ptr = storage;
-    nstored = rd_len;
-    if (nstored > len) {
-	memcpy(buf, store_ptr, len);
-	nreturned += len;
-	store_ptr += len;
-	nstored -= len;
-    } else {
-	memcpy(buf, store_ptr, nstored);
-	nreturned += nstored;
-	nstored = 0;
-    }
-    return(nreturned);
-}
-    
-
-int
-v5_des_write(fd, buf, len)
-     int fd;
-     char *buf;
-     int len;
-{
-  unsigned char *len_buf = (unsigned char *) des_outpkt;
-    
-    if (!do_encrypt)
-      return(write(fd, buf, len));
-    
-    
-    desoutbuf.length = krb5_encrypt_size(len,eblock.crypto_entry);
-    if (desoutbuf.length > sizeof(des_outpkt)-4){
-	syslog(LOG_ERR,"Write size problem.");
-	return(-1);
-    }
-    if ((krb5_encrypt(bsd_context, (krb5_pointer)buf,
-		      desoutbuf.data,
-		      len,
-		      &eblock,
-		      0))){
-	syslog(LOG_ERR,"Write encrypt problem.");
-	return(-1);
-    }
-
-    len_buf[0] = (len & 0xff000000) >> 24;
-    len_buf[1] = (len & 0xff0000) >> 16;
-    len_buf[2] = (len & 0xff00) >> 8;
-    len_buf[3] = (len & 0xff);
-
-    if (write(fd, des_outpkt,desoutbuf.length+4) != desoutbuf.length+4){
-	syslog(LOG_ERR,"Could not write out all data.");
-	return(-1);
-    }
-    else return(len);
-}
-#endif /* KERBEROS */
-
 void usage()
 {
 #ifdef KERBEROS
     syslog(LOG_ERR, 
-	   "usage: klogind [-ke45pP] [-D port] [-w[ip|maxhostlen[,[no]striplocal]]] or [r/R][k/K][x/e][p/P]logind");
+	   "usage: klogind [-ke45pPf] [-D port] [-w[ip|maxhostlen[,[no]striplocal]]] or [r/R][k/K][x/e][p/P]logind");
 #else
     syslog(LOG_ERR, 
-	   "usage: rlogind [-rpP] [-D port] or [r/R][p/P]logind");
+	   "usage: rlogind [-rpPf] [-D port] or [r/R][p/P]logind");
 #endif
 }
 
@@ -1441,9 +1379,11 @@ recvauth(valid_checksum)
     int len;
     krb5_data inbuf;
     char v4_instance[INST_SZ];	/* V4 Instance */
-    char v4_version[9];
+    krb5_data version;
     krb5_authenticator *authenticator;
     krb5_rcache rcache;
+    enum kcmd_proto kcmd_proto;
+    krb5_keyblock *key;
 
     *valid_checksum = 0;
     len = sizeof(laddr);
@@ -1487,8 +1427,8 @@ recvauth(valid_checksum)
 	if (status) return status;
     }
 
-    if ((status = krb5_compat_recvauth(bsd_context, &auth_context, &netf,
-				  "KCMDV0.1",
+    if ((status = krb5_compat_recvauth_version(bsd_context, &auth_context,
+					       &netf,
 				  NULL, 	/* Specify daemon principal */
 				  0, 		/* no flags */
 				  keytab, /* normally NULL to use v5srvtab */
@@ -1502,8 +1442,8 @@ recvauth(valid_checksum)
 
 				  &ticket, 	/* return ticket */
 				  &auth_sys, 	/* which authentication system*/
-				  &v4_kdata, v4_schedule, v4_version))) {
-
+				  &v4_kdata, v4_schedule,
+					       &version))) {
 	if (auth_sys == KRB5_RECVAUTH_V5) {
 	    /*
 	     * clean up before exiting
@@ -1517,7 +1457,25 @@ recvauth(valid_checksum)
 
     getstr(netf, lusername, sizeof (lusername), "locuser");
     getstr(netf, term, sizeof(term), "Terminal type");
-    if ((auth_sys == KRB5_RECVAUTH_V5) && !checksum_ignored) {
+
+    kcmd_proto = KCMD_UNKNOWN_PROTOCOL;
+    if (auth_sys == KRB5_RECVAUTH_V5) {
+	if (version.length != 9) {
+	    fatal (netf, "bad application version length");
+	}
+	if (!memcmp (version.data, "KCMDV0.1", 9))
+	    kcmd_proto = KCMD_OLD_PROTOCOL;
+	else if (!memcmp (version.data, "KCMDV0.2", 9))
+	    kcmd_proto = KCMD_NEW_PROTOCOL;
+    }
+#ifdef KRB5_KRB4_COMPAT
+    if (auth_sys == KRB5_RECVAUTH_V4)
+	kcmd_proto = KCMD_V4_PROTOCOL;
+#endif
+
+    if ((auth_sys == KRB5_RECVAUTH_V5)
+	&& !(checksum_ignored
+	     && kcmd_proto == KCMD_OLD_PROTOCOL)) {
       
       if ((status = krb5_auth_con_getauthenticator(bsd_context, auth_context,
 						   &authenticator)))
@@ -1544,7 +1502,7 @@ recvauth(valid_checksum)
 				      ticket->enc_part2->session->length);
     error_cleanup:
 	if (chksumbuf)
-	    krb5_xfree(chksumbuf);
+	    free(chksumbuf);
 	if (status) {
 	  krb5_free_authenticator(bsd_context, authenticator);
 	  return status;
@@ -1558,14 +1516,14 @@ recvauth(valid_checksum)
 #ifdef KRB5_KRB4_COMPAT
     if (auth_sys == KRB5_RECVAUTH_V4) {
 
-	des_read  = v4_des_read;
-	des_write = v4_des_write;
+	rcmd_stream_init_krb4(v4_kdata->session, do_encrypt, 1, 1);
 
 	/* We do not really know the remote user's login name.
          * Assume it to be the same as the first component of the
 	 * principal's name. 
          */
-	strcpy(rusername, v4_kdata->pname);
+	strncpy(rusername, v4_kdata->pname, sizeof(rusername) - 1);
+	rusername[sizeof(rusername) - 1] = '\0';
 
 	status = krb5_425_conv_principal(bsd_context, v4_kdata->pname,
 					 v4_kdata->pinst, v4_kdata->prealm,
@@ -1584,24 +1542,26 @@ recvauth(valid_checksum)
 				      &client)))
 	return status;
 
-    des_read  = v5_des_read;
-    des_write = v5_des_write;
+    key = 0;
+    status = krb5_auth_con_getremotesubkey (bsd_context, auth_context, &key);
+    if (status)
+	fatal (netf, "Server can't get session subkey");
+    if (!key && do_encrypt && kcmd_proto == KCMD_NEW_PROTOCOL)
+	fatal (netf, "No session subkey sent");
+    if (key && kcmd_proto == KCMD_OLD_PROTOCOL)
+	fatal (netf, "Session subkey not permitted under old kcmd protocol");
+    if (key == 0)
+	key = ticket->enc_part2->session;
+
+    rcmd_stream_init_krb5 (key, do_encrypt, 1, 0, kcmd_proto);
+
+    do_inband = (kcmd_proto == KCMD_NEW_PROTOCOL);
 
     getstr(netf, rusername, sizeof(rusername), "remuser");
 
     if ((status = krb5_unparse_name(bsd_context, client, &krusername)))
 	return status;
     
-    /* Setup up eblock if encrypted login session */
-    /* otherwise zero out session key */
-    if (do_encrypt) {
-	krb5_use_enctype(bsd_context, &eblock,
-			 ticket->enc_part2->session->enctype);
-	if ((status = krb5_process_key(bsd_context, &eblock,
-				       ticket->enc_part2->session)))
-	    fatal(netf, "Permission denied");
-    }      
-
     if ((status = krb5_read_message(bsd_context, (krb5_pointer)&netf, &inbuf)))
 	fatal(netf, "Error reading message");
 
@@ -1613,175 +1573,4 @@ recvauth(valid_checksum)
     return 0;
 }
 
-
-#ifdef KRB5_KRB4_COMPAT
-
-int
-v4_des_read(fd, buf, len)
-int fd;
-register char *buf;
-int len;
-{
-	int nreturned = 0;
-	krb5_ui_4 net_len, rd_len;
-	int cc;
-#if 0
-	unsigned char len_buf[4];
-#endif
-
-	if (!do_encrypt)
-		return(read(fd, buf, len));
-
-	if (nstored >= len) {
-		memcpy(buf, store_ptr, len);
-		store_ptr += len;
-		nstored -= len;
-		return(len);
-	} else if (nstored) {
-		memcpy(buf, store_ptr, nstored);
-		nreturned += nstored;
-		buf += nstored;
-		len -= nstored;
-		nstored = 0;
-	}
-
-#if 0
-	if ((cc = krb_net_read(fd, (char *)len_buf, 4)) != 4) {
-		/* XXX can't read enough, pipe
-		   must have closed */
-		return(0);
-	}
- 	net_len = (((krb5_ui_4)len_buf[0]<<24) |
-		   ((krb5_ui_4)len_buf[1]<<16) |
-		   ((krb5_ui_4)len_buf[2]<<8) |
-		   (krb5_ui_4)len_buf[3]);
-#else
-	{
-	    unsigned char c;
-	    int gotzero = 0;
-
-	    /* We're fetching the length which is MSB first, and the MSB
-	       has to be zero unless the client is sending more than 2^24
-	       (16M) bytes in a single write (which is why this code is in
-	       rlogin but not rcp or rsh.) The only reasons we'd get something
-	       other than zero are:
-	           -- corruption of the tcp stream (which will show up when
-		      everything else is out of sync too)
-		   -- un-caught Berkeley-style "pseudo out-of-band data" which
-		      happens any time the user hits ^C twice.
-	       The latter is *very* common, as shown by an 'rlogin -x -d' 
-	       using the CNS V4 rlogin.         Mark EIchin 1/95
-	      */
-	    do {
-		cc = krb_net_read(fd, &c, 1);
-		if (cc <= 0) return 0; /* read error */
-		if (cc == 1) {
-		    if (c == 0) gotzero = 1;
-		}
-	    } while (!gotzero);
-
-	    if ((cc = krb_net_read(fd, &c, 1)) != 1) return 0;
-	    net_len = c;
-	    if ((cc = krb_net_read(fd, &c, 1)) != 1) return 0;
-	    net_len = (net_len << 8) | c;
-	    if ((cc = krb_net_read(fd, &c, 1)) != 1) return 0;
-	    net_len = (net_len << 8) | c;
-	}
-
-#endif
-	/* Note: net_len is unsigned */
-	if (net_len > sizeof(des_inbuf)) {
-		/* XXX preposterous length, probably out of sync.
-		   act as if pipe closed */
-		return(0);
-	}
-	/* the writer tells us how much real data we are getting, but
-	   we need to read the pad bytes (8-byte boundary) */
-	rd_len = roundup(net_len, 8);
-	if ((cc = krb_net_read(fd, des_inbuf, rd_len)) != rd_len) {
-		/* XXX can't read enough, pipe
-		   must have closed */
-		return(0);
-	}
-	(void) pcbc_encrypt(des_inbuf,
-			    storage,
-			    (net_len < 8) ? 8 : net_len,
-			    v4_schedule,
-			    v4_kdata->session,
-			    DECRYPT);
-	/* 
-	 * when the cleartext block is < 8 bytes, it is "right-justified"
-	 * in the block, so we need to adjust the pointer to the data
-	 */
-	if (net_len < 8)
-		store_ptr = storage + 8 - net_len;
-	else
-		store_ptr = storage;
-	nstored = net_len;
-	if (nstored > len) {
-		memcpy(buf, store_ptr, len);
-		nreturned += len;
-		store_ptr += len;
-		nstored -= len;
-	} else {
-		memcpy(buf, store_ptr, nstored);
-		nreturned += nstored;
-		nstored = 0;
-	}
-	
-	return(nreturned);
-}
-
-int
-v4_des_write(fd, buf, len)
-int fd;
-char *buf;
-int len;
-{
-	static char garbage_buf[8];
-	unsigned char *len_buf = (unsigned char *) des_outpkt;
-
-	if (!do_encrypt)
-		return(write(fd, buf, len));
-
-	/* 
-	 * pcbc_encrypt outputs in 8-byte (64 bit) increments
-	 *
-	 * it zero-fills the cleartext to 8-byte padding,
-	 * so if we have cleartext of < 8 bytes, we want
-	 * to insert random garbage before it so that the ciphertext
-	 * differs for each transmission of the same cleartext.
-	 * if len < 8 - sizeof(long), sizeof(long) bytes of random
-	 * garbage should be sufficient; leave the rest as-is in the buffer.
-	 * if len > 8 - sizeof(long), just garbage fill the rest.
-	 */
-
-#ifdef min
-#undef min
-#endif
-#define min(a,b) ((a < b) ? a : b)
-
-	if (len < 8) {
-		krb5_random_confounder(8 - len, garbage_buf);
-		/* this "right-justifies" the data in the buffer */
-		(void) memcpy(garbage_buf + 8 - len, buf, len);
-	}
-	(void) pcbc_encrypt((len < 8) ? garbage_buf : buf,
-			    des_outpkt+4,
-			    (len < 8) ? 8 : len,
-			    v4_schedule,
-			    v4_kdata->session,
-			    ENCRYPT);
-
-	/* tell the other end the real amount, but send an 8-byte padded
-	   packet */
-	len_buf[0] = (len & 0xff000000) >> 24;
-	len_buf[1] = (len & 0xff0000) >> 16;
-	len_buf[2] = (len & 0xff00) >> 8;
-	len_buf[3] = (len & 0xff);
-	(void) write(fd, des_outpkt, roundup(len,8)+4);
-	return(len);
-}
-
-#endif /* KRB5_KRB4_COMPAT */
 #endif /* KERBEROS */
